import os

#from envpath.settings import *

help_function_dir, this_filename = os.path.split(__file__)
default_feature_dir='.' #The dictory for .bed files, which can be modified as needed
source_script_dir = '.' #The dictory for Rscripts, which can be modified as needed

color_list = [
    "#1f78b4",
    "#ff7f00",
    "#33a02c",
    "#e31a1c",
    "#6a3d9a",
    "#8c564b",
    "#e377c2",
    "#969696",  #
    "#17becf",
    "#253494",
    #'#ffffcc',
    "#a6cee3",
    "#b2df8a",
    "#fb9a99",
    # '#fdbf6f',
    "#cab2d6",
    #'#ffff99',
    "#99d8c9",
    "#8c96c6",
    "#66c2a5",
    "#636363",
    "#ffd92f",
    "#b3b3b3",
    "#db5f57",
]


color_list_2 = [
    "#8dd3c7",
    "#bebada",
    "#fb8072",
    "#80b1d3",
    "#fdb462",
    "#b3de69",
    "#fccde5",
    "#d9d9d9",
    "#bc80bd",
    "#ccebc5",
    "#ffed6f",
    "#ffffb3",
    "#99d8c9",
    "#8c96c6",
    "#66c2a5",
    "#636363",
    "#ffd92f",
    "#b3b3b3",
    "#db5f57",
]
